/*
 * m a i n . c
 *
 * A n t l r  1.0B
 *
 * Terence Parr
 * Purdue University
 * Summer 1989
 *
 * Forced variable names to be unique to 7 char for portability
 * Terence Parr -- March 1990
 */
#include <stdio.h>
#include "set.h"
#include "dlgdef.h"
#include "struct.h"
#include "attrib.h"
#ifdef MEMCHK
#include "trax.h"
#endif

char		**TokenStr = NULL,
			**RegExprStr = NULL,
			**RuleStr = NULL;

set			FiVisit,	/* Track first/follow set visits */
			FoVisit,
			RuleInh,	/* Track whether rule has inherit override */
			*CrossRef;	/* Used to cross ref rules */

int			*RuleFile = NULL;	/* Which file for rule */
int			*RuleLine = NULL;	/* Which line for rule */

char		**ActStr = NULL;	/* Keep all actions before rules */
char		**ActStr2 = NULL;/* Keep all actions after rules */
char		**FailActionStr = NULL;

Rule		Rules;
int			EpToken, EofToken, NumRules=0, NumFiles=0, CurFile=0;
int			MaxToken=0;
int			Save=0;
int			FoSave=0;
int			FoRequest=0;
int			FoCompute=0;
int			FoPartial=0;
int			MaxAlts=15;	/* Updated by grammar to reflect max alts */
int			Terminate=0;

char		*AttribAction=NULL;
char		*FileStr[15] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};

/* Options */
char		*LexFile =	"parser.dlg";
int			ForceInherit = 0;
int			MakeTrax = 0;
int			PrintGrammar = 0;
int			UseCache = 1;
int			NoCode=0;
int			PrFirstSub = 1;
int			PrintFo = 1;
int			PrintFi = 1;
int			CR=0;
int			debug=0;
int			BigErr=0;
int         DfltErr=1;

extern int THRexpr, THRif;

/* O p t i o n   L i s t  S t u f f */
void pDErr()				{ DfltErr = 0; }
void pTrax()				{ MakeTrax = 1; }
void pCache()				{ UseCache = 0; }
void pPrint()				{ PrintGrammar = 1; }
void pNoLex()				{ LexFile = NULL; }
void pLexFile(s,t) char *s,*t;{ LexFile = t; }
void pNoCode()				{ NoCode = 1; pNoLex(); }
void pInherit()				{ ForceInherit = 1; }
void pFile(s) char *s;		{ FileStr[NumFiles++] = s; }
void pPrintSub()			{ PrFirstSub = 0; }
void pPrFi()				{ PrintFi = 0; }
void pPrFo()				{ PrintFo = 0; }
void pCross()				{ CR = 1; }
void pBig()					{ BigErr = 1; }
void pExpr(s,t) char *s,*t;	{ THRexpr = atoi(t); }
void pAlt(s,t) char *s,*t;	{ THRif = atoi(t); }
void pDeb(s,t) char *s,*t;
	{	if ( *t != '.' ) debug |= (1 << (atoi(t)-1));
		else {
			fprintf(stderr, "\t%2s\tPrint follow cache after each follow\n","1");
			fprintf(stderr, "\t%2s\tPrint \"entering function\" kinda stuff\n","2");
			fprintf(stderr, "\t%2s\tPrint guts of follow calculation\n","3");
			fprintf(stderr, "\t%2s\tPrint out cache statistics\n","4");
			fprintf(stderr, "\t%2s\tPrint out memory information at end\n","5");
			exit(0);
		}
	}

typedef struct {
			char *option;
			int  arg;
			void (*process)();
			char *descr;
		} Opt;

Opt options[] = {
	{ "-T", 0, pTrax,	"Place aTrax() call at start of rules" },
	{ "-z", 0, pCache,	"Turn off cache" },
	{ "-p", 0, pPrint,	"Print out the grammar" },
	{ "-n", 0, pNoLex,	"No lexical output (parser.dlg, tokens.h)" },
	{ "-l", 1, pLexFile,"Rename lexical output filename" },
	{ "-c", 0, pNoCode,	"Generate no C code parser" },
	{ "-I", 0, pInherit,"Turn on default inheritance" },
	{ "-s", 0, pPrintSub,"Do not print first sets of subrules" },
	{ "-fi",0, pPrFi,	"Do not print first sets" },
	{ "-fo",0, pPrFo,	"Do not print follow sets" },
	{ "-R", 0, pCross,	"Print rule cross reference" },
	{ "-e", 1, pExpr,	"Set expression threshold (def 3)" },
	{ "-a", 1, pAlt,	"Set alternative threshold (def 3)" },
	{ "-w", 1, pBig,	"Produce better warning messages for subrules" },
	{ "-d", 1, pDeb,	"Turn on debug code (def 0; '.' => list codes)" },
    { "-E", 0, pDErr,   "Turn off \"#define aError(a,b) dfltError(a,b)\"\n" },
	{ "*",  0, pFile, "" },	/* anything else is a file */
	{ NULL, 0, NULL }
};

main(argc, argv)
int argc;
char *argv[];
{
	fprintf(stderr, "Antlr parser generator   Version 1.0B   (c) 1989, 1990\n");
	if ( argc == 1 ) 
	{
		Opt *p = options;
		fprintf(stderr, "antlr [options] f1 f2 ... fn\n");
		while ( *(p->option) != '*' )
		{
			fprintf(stderr, "\t%s %s\t%s\n",
							p->option,
							(p->arg)?"___":"   ",
							p->descr);
			p++;
		}
	}
	ProcessArgs(argc-1, &(argv[1]), options);
	init();
	readDescr();
	if ( Terminate ) { cleanUp(); exit(-1); }
	InitCache(NumRules+1);
	{int new; EpToken = defTname( "[Epsilon]", "e", &new );}
	CrossRef = (set *) calloc(NumRules+1, sizeof(set));
	if ( CrossRef == NULL ) fatal("Cannot allocate cross ref");
	
	/* G e n  O u t p u t */
	if ( LexFile != NULL ) DumpTokens( TokenStr );
	set_size( NumWords(MaxToken) );
	if ( PrintGrammar ) dispRules( Rules, stdout );
	InhChk( Rules );	/* Check inheritance consistency && compute cross ref */
	if ( NoCode ) CheckRules( Rules );
	else GenRules( Rules );
	DEBUG(4) {
		fprintf(stderr, "First cache saved %d Operations\n", Save);
		fprintf(stderr, "There were %d follow set requests\n", FoRequest);
		fprintf(stderr, "%d required no computation\n", FoSave);
		fprintf(stderr, "%d required partial computation\n", FoPartial);
		fprintf(stderr, "%d/%d = %f required complete computation\n",
						FoCompute,FoRequest, ((float)FoCompute)/((float)FoRequest));
		fprintf(stderr, "Computed F(%d rules) = %d complete follows\n", NumRules, FoCompute);
	}

	if ( CR )
	{
		int i;
		
		printf("\nCross Reference:\n\n");
		for (i=1; i<=NumRules; i++)
		{
			printf("Rule %12s referenced by ", RuleStr[i]);
			if ( CrossRef[i].n == 0 ) printf("NO ONE");
			else set_prR(CrossRef[i]);
			putchar('\n');
		}
		printf("\nRules with explicit inheritance: ");
		if ( RuleInh.n == 0 ) printf("NONE");
		else set_prR(RuleInh);
		putchar('\n');
	}

	cleanUp();
	exit(0);
}

static
readDescr()
{
	FILE *f;

	f = NextFile();
	if ( f == NULL )
	{
		warn("\nNo grammar description found (exiting...)");
		cleanUp();
		exit(-1);
	}
	ANTLR(grammar, f);
}

FILE *
NextFile()
{
	FILE *f;

	for (;;)
	{
		if ( CurFile >= NumFiles ) return(NULL);
		f = fopen(FileStr[CurFile], "r");
		if ( f == NULL )
		{
			warn( eMsg("File %s doesn't exist (ignoring...)",
					FileStr[CurFile]) );
			CurFile++;	/* Look for next file */
		}
		else
		{
			aSourceFile = FileStr[CurFile];
			return(f);
		}
	}
}

static
init()
{
	AttribAction = NULL;
	InitStrArray( TokenStr );
	InitStrArray( RegExprStr );
	InitStrArray( RuleStr );
	FiVisit = FoVisit = RuleInh = empty;
}

cleanUp()
{
#ifndef MEMCHK
	if ( AttribAction != NULL ) free( AttribAction );
	set_free( FiVisit );
	set_free( FoVisit );
	set_free( RuleInh );
	if ( RuleFile != NULL ) free( RuleFile );
	if ( RuleLine != NULL ) free( RuleLine );
	if ( Rules != NULL ) freeRules( Rules );
	if ( TokenStr != NULL ) freeStrArray( TokenStr );
	if ( RegExprStr != NULL ) freeStrArray( RegExprStr );
	if ( ActStr != NULL ) freeStrArray( ActStr );
	if ( ActStr2 != NULL ) freeStrArray( ActStr2 );
	if ( FailActionStr != NULL ) freeStrArray( FailActionStr );
	if ( RuleStr != NULL ) freeStrArray( RuleStr );
	if ( CrossRef != NULL )
	{
		int i;
		for (i=1; i<=NumRules; i++)
		{
			set_free( CrossRef[i] );
		}
		free( CrossRef );
	}
	KillCache(NumRules+1);
#else
	mWipe();		/* Free all memory */
	DEBUG(5) {
		mHist(stderr);
		fprintf(stderr, "Max memory allocated at one time, %d\n", mMax);
	}
#endif
}

char *
eMsg(s, a1, a2, a3, a4, a5, a6)
char *s;
long a1,a2,a3,a4,a5,a6;
{
	static char buf[150];
	
	sprintf(buf, s, a1, a2, a3, a4, a5, a6);
	return( buf );
}

/*
 * Add rule 'r' to the RuleStr array if not already there.
 * Return position where 'r' was placed.
 * Return 0 if duplicate.
 */
defRname(r)
char *r;
{
	static int i=0;
	char *p;
	
	if ( strPos(RuleStr, r) != 0 )
	{
		warn( eMsg("Duplicate rule redefinition: %s", r) );
		return(0);
	}
	
	p = (char *) calloc(strlen(r)+1, sizeof(char));
	if ( p == NULL ) fatal("Can't Alloc RuleName");
	strcpy(p, r);
	extStrArray(RuleStr, i);
	RuleStr[i] = p;
	extStrArray(RuleStr, i);
	RuleStr[i] = NULL;
	return( --i );
}

/*
 * Add token 't' to the TokenStr array if not already there and a token name.
 * If reg expr, add to reg expr array
 * Return position where 't' was placed.
 */
defTname(t, e, new)
char *t, *e;
int *new;
{
	int i=0,j=0;
	char *p;
	
/*
	fprintf(stderr, "entering defTname, %s == %s\n",
						(t==NULL)?"NULL":t,
						(e==NULL)?"NULL":e);
	fprintf(stderr, "Tokens\n");
	for (k=1; k<sizeT(TokenStr); k++)
	{
		fprintf(stderr, "[%d] '%s'\n", k, (TokenStr[k]==NULL)?"NULL":TokenStr[k]);
	}
	fprintf(stderr, "\nExprs\n");
	for (k=1; k<sizeT(RegExprStr); k++)
	{
		fprintf(stderr, "[%d] '%s'\n", k, (RegExprStr[k]==NULL)?"NULL":RegExprStr[k]);
	}
*/

	if ( e == NULL && t == NULL ) fatal("defTname: Hmmm... both ptrs are NULL");
	if ( t != NULL ) i=strPos(TokenStr, t);
	if ( e != NULL ) j=strPos(RegExprStr, e);
	if ( i != j && i!=0 && j!=0 )
	{
		warnFL( eMsg("Token name %s and rexpr %s are different tokens; ignored",
						TokenStr[i], RegExprStr[j]),
				aSourceFile,
				lex_line );
	}
	if ( i!=0 || j!=0 ) *new = FALSE;
	else *new = TRUE;
	if ( i==0 && t != NULL )
	{
		p = (char *) calloc(strlen(t)+1, sizeof(char));
		if ( p == NULL ) fatal("Can't Alloc TokenName");
		strcpy(p, t);
		if ( j == 0 )
		{
			/* doesn't exist anywhere, create the token */
			extStrArray(TokenStr, MaxToken); TokenStr[MaxToken--] = p;
			extStrArray(RegExprStr, MaxToken);
			i = MaxToken;
		}
		else
		{
			if ( TokenStr[j] != NULL )
			{
				warnFL( eMsg("Redefinition of token name %s for rexpr %s to %s; ignored",
								TokenStr[j], RegExprStr[j], p),
						aSourceFile,
						lex_line );
				free(p);
			}
			else TokenStr[j] = p;
		}
	}
	if ( j==0 && e != NULL )
	{
		p = (char *) calloc(strlen(e)+1, sizeof(char));
		if ( p == NULL ) fatal("Can't Alloc RegExprStr");
		strcpy(p, e);
		if ( i == 0 )
		{
			/* must create token */
			extStrArray(RegExprStr, MaxToken); RegExprStr[MaxToken--] = p;
			extStrArray(TokenStr, MaxToken);
			j = MaxToken;
		}
		else
		{
			if ( RegExprStr[i] != NULL )
			{
				warnFL( eMsg("Redefinition of rexpr %s for token name %s to %s; ignored",
								RegExprStr[i], TokenStr[i], p),
						aSourceFile,
						lex_line );
				free(p);
			}
			else RegExprStr[i] = p;
		}
	}

	return( (i!=0)?i:j );
}

freeAtom(a)
Atom *a;
{
	if ( a == NULL ) return;
	switch ( a->type )
	{
		case aRule :
			free( a->ptr.rule );
			if ( a->inheritance != NULL ) free( a->inheritance );
			break;
		case aAction :
			free( a->ptr.action );
			break;
		case aSubBLK :
		case aOptBLK :
		case aLoopBLK :
			freeBlk( a->ptr.block );
			if ( a->inheritance != NULL ) free( a->inheritance );
			break;
	}
}

freeAlt(alt)
Atom *alt;
{
	int i;
	
	if ( alt == NULL ) return;
	for (i=1; alt[i].type != EOL; i++) freeAtom( &(alt[i]) );
	free( alt );
}

freeBlk(b)
Atom **b;
{
	int i;
	
	if ( b == NULL ) return;
	for (i=1; b[i] != NULL; i++) freeAlt( b[i] );
	free( b );
}

freeRules(r)
Atom ***r;
{
	int i;
	
	for (i=1; r[i] != NULL; i++) freeBlk( r[i] );
	free( r );
}

strPos(a, s)
char **a, *s;
{
	int i;
	
	for (i=1; i<sizeT(a); i++)
	{
		if ( a[i]==NULL ) continue;
		if ( strcmp(a[i],s) == StrSame ) return( i );
	}
	return( 0 );
}

freeStrArray(a)
char **a;
{
	int i;
	
	for (i=1; i<sizeT(a); i++)
	{
		if ( a[i] != NULL )
			if ( a[i]!=NULL ) free( a[i] );
	}
	free( a );
}

dispAtom(a,f)
Atom a;
FILE *f;
{
	switch ( a.type )
	{
		case aLeaf :	fprintf(f," %s", TokenName(a.token));
						break;
		case aRule :	fprintf(f," %s", a.ptr.rule);
						break;
		case aAction :	/*fprintf(f,"<<...>>");*/
						break;
		case aSubBLK :	fprintf(f," (");
						dispBlk(a.ptr.block,f);
						fprintf(f," )");
						break;
		case aOptBLK :	fprintf(f," {");
						dispBlk(a.ptr.block,f);
						fprintf(f," }");
						break;
		case aLoopBLK :	fprintf(f," (");
						dispBlk(a.ptr.block,f);
						fprintf(f," )*");
						break;
	}
}

dispAlt(alt,f)
Atom *alt;
FILE *f;
{
	int i;
	
	for (i=1; alt[i].type != EOL; i++)
	{
		dispAtom( alt[i], f );
	}
}

dispBlk(b,f)
Atom **b;
FILE *f;
{
	int i;
	
	for (i=1; b[i] != NULL; i++)
	{
		dispAlt( b[i], f);
		if ( b[i+1]!=NULL ) fprintf(f," |");
	}
}

dispRules(r,f)
Atom ***r;
FILE *f;
{
	int i, old;
	
	old = -1;
	for (i=1; r[i] != NULL; i++)
	{
		if ( RuleFile[i] != old )
		{
			old = RuleFile[i];
			fprintf(f,"\n[File %s]\n\n", FileStr[old]);
		}
		fprintf(f,"%s :", RuleStr[i]);
		dispBlk( r[i],f );
		fprintf(f," ;\n\n");
	}
}

ProcessArgs(argc, argv, options)
int argc;
char **argv;
Opt *options;
{
	Opt *p;
	
	while ( argc-- > 0 )
	{
		p = options;
		while ( p->option != NULL )
		{
			if ( strcmp(p->option, "*") == 0 ||
				 strcmp(p->option, *argv) == 0 )
			{
				if ( p->arg )
				{
					(*p->process)( *argv, *(argv+1) );
					argv++;
					argc--;
				}
				else
					(*p->process)( *argv );
				break;
			}
			p++;
		}
		argv++;
	}
}
